program Stut;
{-------------------------------------------------------------
{ W. J. Hansen
{ Copyright, 1981, Three Rivers Computer Company
{
{ Abstract:
{   Main program for streamer utility.
{-------------------------------------------------------------}

{-------------------------------------------------------------
{ Change Log:
{  4 jun 83  RDB   V1.8   changes to tsio.micro
{
{  11 Feb 83  WJH  V1.7   V2.8 of TSInterface.
{
{   7 Feb 83  WJH  V1.6   V2.7 of TSInterface and TSIO.
{  
{  25 Jan 83  WJH  V1.5   V2.6 of TSInterface and modifications to TSIO.
{
{  15 Dec 82  WJH  V1.4   Add confirmation request to erase command.
{
{  15 Dec 82  WJH  V1.3   Add Clear command.  Forbid duplicate names on tape.
{                         Warn that directory name not put on tape.
{                         Properly reset LastSel when not continue after 
{                         selection error.  V2.5 micro code.
{                         Use ConvUpper instead of CnvUpper  (in Stut itself).
{                         Add beeps for warnings in DoPut and DoGet.
{  
{  23 Nov 82  WJH  V1.2   fixes to STFOWrite.
{
{  23 Nov 81  WJH  V1.1   changes to STFOWrite and TSInterface
{
{  29 Oct 81  WJH  V1.0   Initial release.
{
{  28 Oct 81  WJH  V0.9   Remove : and > from tape partition names
{                         Add BUFFERS and @ to help information
{                         Change final message to mention "after rewind"
{
{  26 Oct 81  WJH  V0.8   COnvert tape part names to upper case for comparisons
{
{  10 Sep 81  WJH  V0.7   only give 'remove cartridge' at final exit
{                         REWIND use TSPosn
{
{   9 Sep 81  WJH  V0.6   remove period in 'Initial status messages'
{                         for Quit give reminder to remove cartridge
{                         REWIND use TSReset
{                         only CnvUpper the Cmd itself, not whole command line
{
{   8 Sep 81  WJH  V0.6   make BufferSize command
{                         use WriteLn when echoing commands from file
{
{   2 Sep 81  WJH  V0.5   make DoGet and DoPut handle arbitrary directories
{                         use ReadToFM in DoGet to skip tape header
{                         change DIRECT to list file names
{                         add HEADER command to list tape header
{                         add FSData and NumInUse to PartBlk
{                         DoGet and DoPut pause for directory errors
{
{   1 Sep 81  WJH  V0.4   put in SelDir and SelTpPart
{                         put in PartBlk 
{                         add STUTTitle
{
{  28 Aug 81  WJH  V0.3   call RelSTBuffers
{                         writeln of command error messages
{
{  19 Aug 81  WJH  V0.2   Add Direct, implement Get
{
{  18 Aug 81  WJH  V0.1   Add Help, List
{                         Implement Select, Put
{    
{  17 Aug 81  WJH  V0.0   Create file
{-------------------------------------------------------------}


EXPORTS


const
   STVersion = 'V1.8';
   TapeFormat = 'T1.2';

var
   STUTTitle: String;

imports FileDefs from FileDefs;      {for PathName}
imports GetTimeStamp from GetTimeStamp;

type 
   ptrSelectEntry = ^SelectEntry;
   SelectEntry = record
         SelSrc, SelDest: PathName;
         SelTpPart: SimpleName;
         SelDir: PathName;
         SelAddr: DiskAddr;
         SelNext: ptrSelectEntry;
   end;
   
   ptrTapeHeader = ^TapeHeader;
   TapeHeader = packed record
         TpId, TpFormat, TpSTVers: packed array[0..3] of char;
         TpTime: TimeStamp;
         TpWriter: String[31];
         TpMachine: String[25];
         TpNote: String[151];
         TpNBootBlks: integer;
         TpNParts:    integer;
         TpPartName: array [1..10] of String[25];
         TpNDirBlks: integer;
         TpUnUsed: String[19];
   end;
   
   ptrPartBlk = ^PartBlk;
   PartBlk = packed record
         PartId: packed array [0..3] of char;
         PartName:  SimpleName;
         PartOriginal: PathName;
         PartFSData: FSDataEntry;
         PartNumInUse: integer;
   end;

procedure LogError(Msg: String);   


PRIVATE

imports System      from System;            {for UsrCmdLine}
imports Stream      from Stream;            {for PastEof}
imports IO_Unit     from IO_Unit;           {for IOBeep}
imports FileUtils   from FileUtils;         {for FSAddToTitle}
imports FileDir     from FileDir;           {for DefaultDeviceName}
imports AllocDisk   from AllocDisk;
imports CmdParse    from CmdParse;
imports Perq_String from Perq_String;
imports Clock       from Clock;

imports STFOWrite   from STFOWrite;
imports STFORead    from STFORead;
imports TSInterface from TSInterface;
imports STBufUts    from STBufUts;
imports STBufWrite  from STBufWrite;
imports STBufRead   from STBufRead;


var
   CmdTable : CmdArray;

const
   SelectCmd = 1;
   ListCmd = SelectCmd+1;
   ClearCmd = ListCmd+1;
   PutCmd = ClearCmd+1;
   GetCmd = PutCmd+1;
   HeadCmd = GetCmd+1;
   DirCmd = HeadCmd+1;
   RewindCmd = DirCmd+1;
   WindCmd = RewindCmd+1;
   EraseCmd = WindCmd+1;
   BufSzCmd = EraseCmd+1;
   HelpCmd = BufSzCmd+1;
   QuitCmd = HelpCmd+1;
   
   NCmds = QuitCmd;
   IGLCmd = NCmds+1;
   AmbigCmd = NCmds+2;

procedure InitCmds;
begin
   CmdTable[SelectCmd] := 'SELECT';
   CmdTable[ListCmd]   := 'LIST';
   CmdTable[ClearCmd]  := 'CLEAR';
   CmdTable[PutCmd]    := 'PUT';
   CmdTable[GetCmd]    := 'GET';
   CmdTable[HeadCmd]   := 'HEADER';
   CmdTable[DirCmd]    := 'DIRECT';
   CmdTable[RewindCmd] := 'REWIND';
   CmdTable[WindCmd]   := 'WIND';
   CmdTable[EraseCmd]  := 'ERASE';
   CmdTable[BufSzCmd]  := 'BUFFERSIZE';
   CmdTable[HelpCmd]   := 'HELP';
   CmdTable[QuitCmd]   := 'QUIT';
end;

procedure GiveHelp;
begin
   writeln('Commands available are');
   writeln('   SELECT old-partition-name [new-partition-name]');
   writeln('           Selects the named partitions for transfer');
   writeln('           and specifies there names on the destination.');
   writeln('   LIST    Lists the selected partitions.');
   writeln('   CLEAR   Removes all existing selections.');
   writeln('   PUT     Transfers selected partitions to tape.');
   writeln('   GET     Transfers selected partitions to disk.');
   writeln('   @filename      Takes commands from named file.');
   writeln('   HEADER  Displays contents of tape header.');
   writeln('   DIRECT  Scans tape and displays names of all files.');
   writeln('   REWIND  Rewinds the tape.');
   writeln('   WIND    Moves to end of tape and back to beginning.');
   writeln('   ERASE   Erases the tape.');
   writeln('   BUFFERS nn     Sets number of non-screen buffers.');
   writeln('   HELP    Prints this text.');
   writeln('   QUIT    Leaves the program.');
end;


function Ask(msg, def: string) : boolean;
var
   s:string; gotanswer:boolean;
begin
   write(msg);
   if def <> '' then Write('[',def,'] ');
   gotanswer := false;
   repeat
      readln(s);
      if length(s) = 0 then s := def;
      if length(s) > 0 then
         gotanswer := s[1] in ['y','Y','n','N'];
      if not gotanswer then write('Yes or No? ')
   until gotanswer;
   ask := s[1] in ['y','Y']
end;



procedure LogError(Msg: String);
{------------------------------------------
{ Abstract:
{    Log an error message.  
{
{ Parameters:
{    Msg - A description of the error.
{ Note:
{    Eventually LogError should write errors into a file.
{    It might also accept a severity code to indicate
{    whether it ought to pause until user sees the error.
{------------------------------------------}
begin
   writeln;
   writeln(' ** ** ', Msg, ' ** ** ');
   IOBeep;
end;


function RdTpHdr: ptrTapeHeader;
var
   TpHdr: ptrTapeHeader;
   Status: StatBuf;
   BufPtr: BufList;
   Got: integer;
begin
   new (TpHdr);
   BufPtr[1] := Recast(TpHdr, PtrDiskBuffer);
   CallRdSetup(Status);
   Got := 1;
   if CallRdBuffers(Got, BufPtr, 1) then {must not fail};
   if Got<1 then raise TSBufError(5);
   CallRdEnd(Status);
   RdTpHdr := TpHdr;
end;


var 
   Selections, LastSel : ptrSelectEntry;
   
procedure DoSelection(Names: CString);
var
   Break: Cstring;
   TSel: ptrSelectEntry;
begin
   new(TSel);
   with TSel^ do begin
      RemDelimiters(Names, ' ', Break);
      GetSymbol(Names, SelSrc, ' ', Break);
      RemDelimiters(Names, ' ', Break);
      if length(Names) > 0 then
         GetSymbol(Names, SelDest, ' ', Break)
      else SelDest := SelSrc;
      SelNext := nil;
      LastSel^.SelNext := TSel;     {append new one at end of list}
      LastSel := TSel;
   end;
end;

procedure DoList;
var
   TSel : ptrSelectEntry;
begin
   TSel := Selections^.SelNext;
   writeln('   Selected partitions are');
   while TSel <> nil do begin
      write ('      Src: ', TSel^.SelSrc);
      writeln ('    Dest: ', TSel^.SelDest);
      TSel := TSel^.SelNext;
   end;
end;

procedure DoClear;
begin
   DoList;
   if ask('Erase selections? ', 'yes') then begin
       Selections^.SelNext := nil;
       LastSel := Selections;
   end;
end;


var   
   BufferSize: integer;
   
   
procedure DoBufSz(S: CString);
var
   Digit: integer;
   Val: integer;
   Break: CString;
begin
   RemDelimiters(S, ' ', Break);
   Val := 0;
   while length(S) > 0 do begin
      Digit := ord(S[1])-ord('0');
      if (0<=Digit) and (Digit<=9) then begin
         Val := 10*Val + Digit;
         S := substr(S, 2, length(S)-1);
      end
      else S := '';
   end;
   BufferSize := Val;
end;   


procedure DoGet;
var
   TSel : ptrSelectEntry;
   Okay: Boolean;
   i: integer;
   PartNo: integer;
   DiskPartName: string;
   TpPtName: string;
   Status: StatBuf;
   IgnoreInt: integer;
   BufPtr: ptrDiskBuffer;
   TpHdr: ptrTapeHeader;
   tName: string;
begin
   TpHdr := RdTpHdr;
   TSel := Selections;
   while TSel^.SelNext <> nil do with TSel^.SelNext^ do begin
      Okay := True;
      
      {build PathName of directory}
      if SelDest[length(SelDest)]<>'>' then
         AppendChar(SelDest, '>');    
      i := PosC(SelDest, ':');
      if i=1 then SelDest := substr(SelDest, 2, length(SelDest)-1);
      if i <= 1 then SelDest := concat(DefaultDeviceName, SelDest);
      i := PosC(SelDest, '>');
      DiskPartName := Substr(SelDest, 1, i);
      PartNo := FindPartition(DiskPartName);
      if PartNo = 0 then begin
         writeln ('No partition for: ', SelDest);
         Okay := False;
      end
      else if not PartTable[PartNo].PartInUse then begin
         writeln ('Partition not in use: ', SelDest);
         Okay := False;
      end;
      if Okay then begin
         i := PosC(SelDest, '>');
         if i < length(SelDest) then
            SelAddr := FileIdToSegID (FSLocalLookup (
                    Concat (Substr(SelDest, 1, length(SelDest)-1), '.DR'),
                    IgnoreInt, IgnoreInt))
         else
            SelAddr :=  PartTable[PartNo].PartRootDir;
         if SelAddr=DBLZERO then begin
            Okay := False;
            writeln('Destination directory not found', SelDest);
         end;
      end;
      if Okay then begin
         SelDir := SelDest;
         i := PosC(SelSrc, ':');
         if i > 0 then 
            SelSrc := Substr(SelSrc, i+1, length(SelSrc)-i);
         i := PosC(SelSrc, '>');
         if i > 0 then begin
            if i<length(SelSrc) then begin
                tName := substr(SelSrc, i, length(SelSrc)-i+1);
                IOBeep;
                writeln('Tape partitions do not have directory names. ',
                           ' Ignoring: "', tName, '"');
            end;
            SelSrc := Substr(SelSrc, 1, i-1);
         end;
         if Length(SelSrc) > 25 then begin
            SelSrc := substr(SelSrc, 1, 25);
            IOBeep;
            writeln('Source name truncated to 25 chars: ',
                     SelSrc);
         end;
         SelTpPart := SelSrc;
         ConvUpper(SelTpPart);
         Okay := false;
         for i := 1 to TpHdr^.TpNParts do begin
            TpPtName := TpHdr^.TpPartName[i];
            ConvUpper(TpPtName);
            if TpPtName = SelTpPart then Okay := True;
         end;
         if not Okay then 
            writeln('No tape partition named ', SelSrc);
      end;
      if Okay then begin
         writeln('Read partition: ', SelSrc, 
                 ' from tape into: ', SelDest);
         TSel := TSel^.SelNext;
      end
      else begin
         TSel^.SelNext := SelNext;   {delete current block from chain}
         IOBeep;
         if not ask('Continue? ', 'yes') then begin
            LastSel := Selections;
            while LastSel^.SelNext <> NIL do
                LastSel := LastSel^.SelNext;
            exit(DoGet);
         end;
      end;
   end;
   
   if Selections^.SelNext <> nil then 
      FORead(Selections^.SelNext, BufferSize);
   
   Selections^.SelNext := nil;
   LastSel := Selections;
end;


procedure DoPut(Note: CString);
var
   TSel : ptrSelectEntry;
   Got: BufRange;
   Okay: Boolean;
   i: integer;
   PartNo: integer;
   DiskPartName: String;
   TpHdr: TapeHeader;
   Status: StatBuf;
   Break: String[1];
   IgnoreInt: integer;
   dName, tName: string;
begin
   RemDelimiters(Note, ' ', Break);
   if length(Note) = 0 then begin
      writeln('Annotation for tape header:');
      readln(Note);
   end;
   TpHdr.TpNParts := 0;
   TSel := Selections;
   while TSel^.SelNext <> nil do with TSel^.SelNext^ do begin
      Okay := True;

      {build PathName of directory}
      if SelSrc[length(SelSrc)]<>'>' then
         AppendChar(SelSrc, '>');    
      i := PosC(SelSrc, ':');
      if i=1 then SelSrc := substr(SelSrc, 2, length(SelSrc)-1);
      if i <= 1 then SelSrc := concat(DefaultDeviceName, SelSrc);
      i := PosC(SelSrc, '>');

      DiskPartName := Substr(SelSrc, 1, i);
      PartNo := FindPartition(DiskPartName);
      if PartNo = 0 then begin
         writeln ('No partition for: ', SelSrc);
         Okay := False;
      end
      else if not PartTable[PartNo].PartInUse then begin
         writeln ('Partition not in use: ', SelSrc);
         Okay := False;
      end;
      if Okay then begin
         i := PosC(SelSrc, '>');
         if i < length(SelSrc) then
            SelAddr := FileIdToSegID (FSLocalLookup (
                    Concat (Substr(SelSrc, 1, length(SelSrc)-1), '.DR'),
                    IgnoreInt, IgnoreInt))
         else
            SelAddr :=  PartTable[PartNo].PartRootDir;
         if SelAddr=DBLZERO then begin
            Okay := False;
            writeln('Source directory not found', SelDest);
         end;
      end;
      if Okay then begin
         i := PosC(SelDest, ':');
         if i > 0 then 
            SelDest := Substr(SelDest, i+1, length(SelDest)-i);
         i := PosC(SelDest, '>');
         if i > 0 then begin
            if i<length(SelDest) then begin
                dName := substr(SelDest, i, length(SelDest)-i+1);
                IOBeep;
                writeln('Only partition names allowed on tape.  Ignoring "', 
                             dName, '"');
            end;
            SelDest := Substr(SelDest, 1, i-1);
         end;
         if Length(SelDest) > 25 then begin
            SelDest := substr(SelDest, 1, 25);
            IOBeep;
            writeln('Destination name truncated to 25 chars: ',
                     SelDest);
         end;
         {check that partition name is unique on tape}
         dName := SelDest;
         convUpper(dName);
         for i := 1 to TpHdr.TpNParts do begin
             tName := TpHdr.TpPartName[i];
             convupper(tName);
             if tName = dName then begin
                 Okay := false;
                 writeln('Duplicate names not allowed on tape: ', SelDest);
             end;
         end;
      end;
      if Okay then begin
         writeln('Put directory ', SelSrc, 
                 ' on tape as :', SelDest);
         SelTpPart := SelDest;
         SelDir := SelSrc;
         TpHdr.TpNParts := TpHdr.TpNParts+1;
         TpHdr.TpPartName[TpHdr.TpNParts] := SelDest;
         TSel := TSel^.SelNext;
      end
      else begin
         TSel^.SelNext := SelNext;   {delete current block from chain}
         IOBeep;
         if not ask('Continue? ', 'yes') then begin
            LastSel := Selections;
            while LastSel^.SelNext <> NIL do
                LastSel := LastSel^.SelNext;
            exit(DoPut);
         end;
      end;
   end;
   
   with TpHdr do begin
      TpId     := 'StUt';
      TpFormat := TapeFormat;
      TpSTVers := STVersion;
      GetTStamp(TpTime);
      TpWriter := CurUserName;
      TpMachine := '                         ';
      TpNote := Note;
      TpNBootBlks := 0;
      TpNDirBlks := 0;
   end;
   
   if Selections^.SelNext <> nil then begin
      InitSTBuffers(BufferSize);
      i := AllocWrBuffers(3, Got);
      if Got<>3 then Raise TSBufError(0);
      BufPtrs[i]^   := Recast(TpHdr, DiskBuffer);
      BufPtrs[i+1]^ := Recast(TpHdr, DiskBuffer);
      BufPtrs[i+2]^ := Recast(TpHdr, DiskBuffer);
      DoneWrBuffers(3, FileMark);
      
      CallWrSetup(Status);
      PrtStatus(Status, 'Initial status before writing');
      FlushWrBuffers;

      FOWrite(Selections^.SelNext);
      
      RelSTBuffers;
   end;
   
   Selections^.SelNext := nil;
   LastSel := Selections;
end;


procedure DoHeader;
var
   TpHdr: ptrTapeHeader;
   Time:  TimeString;
   i: integer;
begin
   TpHdr := RdTpHdr;
   with TpHdr^ do begin
      writeln('Tape: ', TpNote);
      StampToString(TpTime, Time);
      writeln('Written: ', Time, '    By: ', TpWriter, 
                    '        (Format: ', TpFormat, ')');
      write ('Partitions:   ');
      for i := 1 to TpNParts do
         write (TpPartName[i]:14);
      writeln;
   end;
end;

   
procedure DoTapePosn(Command: integer);
var Status: StatBuf;
begin
   if Command = #042 {ERASE} then
       if not ask('Do you really want to ERASE all tape data? ', 'No') then
           exit(DoTapePosn);
   CallTSPosn(Command, Status);
   PrtStatus(Status, 'Status when starting tape movement');
end {DoTapePosn};






procedure DoCmds(var F: Text;  Console: Boolean);  forward;

procedure DoCmd (CmdLine: CString;  var Continue: Boolean);
label 30;
handler TSError(Status:StatBuf; Msg: TSMsg);
begin
   PrtStatus(Status, Msg);
   goto 30;
end;
var
   Cmd, Break : CString;
   G : Text;
   Status: StatBuf;
   FileName: PathName;
   IgnoreBlocks, IgnoreBits: integer;
begin
   if length(CmdLine)=0 then {nothing}
   else if CmdLine[1]='@' then begin
      RemDelimiters(CmdLine, '@', Break);
      FileName := CmdLine;
      ConvUpper(CmdLine);
      if CmdLine = 'CONSOLE:' then begin
         Reset(G, CmdLine);
         DoCmds(G, True);
         close(G);
      end
      else if FSExtSearch(FSSysSearchList, ' .Cmd ', FileName, 
                          IgnoreBlocks, IgnoreBits)        = 0 then
         LogError(concat('No such command file as ', FileName))
      else begin
         Reset (G, FileName);
         DoCmds (G, False);
         close(G);
      end;
   end 
   else begin
      GetSymbol(CmdLine, Cmd, ' ', Break);
      ConvUpper(Cmd);
      case UniqueCmdIndex(Cmd, CmdTable, NCmds) of
         SelectCmd: DoSelection(CmdLine);
         ListCmd: DoList;
         ClearCmd: DoClear;
         PutCmd: DoPut(CmdLine);
         GetCmd: DoGet;
         HeadCmd: DoHeader;
         DirCmd:begin
                   DoHeader;
                   TapeDirectory(BufferSize);
                end;
         RewindCmd: DoTapePosn(#041 {REWIND});
         WindCmd: DoTapePosn(#044 {RETENSION});
         EraseCmd: DoTapePosn(#042 {ERASE});
         BufSzCmd: DoBufSz(CmdLine);
         HelpCmd: GiveHelp;
         QuitCmd: Continue := False;
         IGLCmd:  writeln('Command "', Cmd, '" is unknown.  ',
                        'Type HELP if you need help.');
         AmbigCmd: writeln('Command "', Cmd, '" is ambiguous.  ',
                          'Type a longer word.');
      end;
   end;
30:
end {DoCmd};


procedure DoCmds(var F: Text;  Console: Boolean);
var 
   Stutting : Boolean;
   CmdLine : CString;
label 30;
handler PastEof(FileName:PathName);
begin
   goto 30;
end;
begin
   Stutting := True;
   while Stutting do begin
      if Console then write('>');
      readln(F, CmdLine);
      if not Console then writeln ('@>', CmdLine);
      DoCmd(CmdLine, Stutting);
   end;
30:
end;







{ S T U T   starts here.     }


                   
label 30;
handler TSBufError(Code: integer);
begin
   writeln('TSBufError: ', Code:1);
   goto 30;
end;
         
handler TSError(Status:StatBuf; Msg: TSMsg);
begin
   PrtStatus(Status, Msg);
   goto 30;
end;

var
   Cmd, Break: CString;
   DummyBoolean: Boolean;

begin {STUT}
   STUTTitle := Concat ('  STUT ', STVersion);
   FSAddToTitleLine (STUTTitle);
   InitCmds;
   TSInit;
   BufferSize := 120;
   new(Selections);
   Selections^.SelNext := nil;
   LastSel := Selections;
   GetSymbol(UsrCmdLine, Cmd, ' ', Break);
   RemDelimiters(UsrCmdLine, ' ', Break);
   if Length(UsrCmdLine) > 0 then
      DoCmd(UsrCmdLine, DummyBoolean)
   else DoCmds(input, True{is console});
30:
   writeln('*** ***   Remove tape when rewind is done.   *** ***');
end {STUT}.
