module STFOWrite;
{------------------------------------------------
{ Abstract:
{   Writes to streamer tape in file order.
{
{ WJHansen
{ Copyright 1981, Three Rivers Computer Corporation
{------------------------------------------------}
{------------------------------------------------
{ Change Log:
{
{ 23 Nov 82 WJH V1.3   fix using wrong buffer to print partition name
{
{ 12 May 82 WJH V1.2   fix TSbuferror(0) bug
{
{ 23 Nov 81 WJH V1.1   check name length in DumpFile
{
{ 28 Oct 81 WJH V1.0   Put in Documentation headers.
{
{ summer 81 WJH V0.0   Create file
{------------------------------------------------}
exports

imports Stut from Stut;      {for ptrSelectEntry}
   
procedure FOWrite(Selections: ptrSelectEntry);

   

private                        

   imports TSInterface from TSInterface;
   imports STBufUts    from STBufUts;
   imports STBufWrite  from STBufWrite;

   imports DiskIO      from DiskIO;
   imports AllocDisk   from AllocDisk;
   imports FileDefs    from FileDefs;
   imports FileTypes   from FileTypes; 
   imports FileUtils   from FileUtils;
   imports Memory      from Memory;
   imports Clock       from Clock;
   imports Perq_String from Perq_String;

var 
   TotalBlocks, TotalFiles, TotalErrors: integer;
                           {Totals for data in partition.}
   GrandTBlocks, GrandTFiles, GrandTErrors: integer;
                           {Grand totals}
   GlobalFIB: ptrDiskBuffer;
   GlobalHeader: ptrHeader;
   NameStart: integer;    {position of first char of file names after
                          directory selected for PUTting to tape.}
    


function ReadABlock (Addr: DiskAddr; Buff: ptrDiskBuffer; Hd: ptrHeader):
                                                              Boolean;
{------------------------------------------------
{ Abstract:
{   Reads a block from disk.  Counts errors, if any.
{ Parameters:
{   DiskAddr - where to read from
{   Buff     - where to put block of data
{   Hd       - where to put label
{ Returns:
{   True if no error.  False if errror.
{------------------------------------------------}
begin
   ReadABlock := False;
   if not TryDiskIO(Addr, Buff, Hd, DskRead, NUMTRIES)  then begin
      write ('Read error.');
      TotalErrors := TotalErrors+1;
   end
   else 
      ReadABlock := True;
end {ReadABlock};



procedure DumpFile(CurrBlkAddr: DiskAddr); 
{------------------------------------------------
{ Abstract:
{   Writes a file to streamer tape.
{ Parameters:
{   CurrBlkAddr - Initially the address of block zero of the file.
{         Is later set to successive block locations.
{         (Is not a var parameter, so caller is unaffected.)
{ Environment:
{   GlobalFIB must point to an area containing the FIB for the file.
{------------------------------------------------}
var
   BufAddr: ptrDiskBuffer;
   Got: BufRange;
   L: integer;
begin
   TotalFiles := TotalFiles+1;
{      FileSparse is not currently set by system}
{  if GlobalFIB^.FSData.FileSparse then         }
{     LogError('Writing a sparse file');        }
   {Remove superordinate directory names from FSData}
   L := Length(GlobalFIB^.FSData.FileName)-NameStart+1;
      {The following tests of L are included because the name
      "Floppy.Scratch.A" has occasionally caused a length error.}
      {Note that "80" is the size of PartialPathName, the type of
      GlobalFIB^.FSData.FileName.}
   if L <= 0 then L := 0
   else if L > 80 then L := 80;
   GlobalFIB^.FSData.FileName :=
          substr(GlobalFIB^.FSData.FileName, NameStart, L);
   BufPtrs[ AllocWrBuffers(1, Got) ]^ := GlobalFIB^;
   DoneWrBuffers(1, NotEof);
   while CurrBlkAddr <> DBLZERO do  begin
         TotalBlocks := TotalBlocks + 1;
         BufAddr := BufPtrs[ AllocWrBuffers(1, Got) ];
         {Read Block CurrBlkAddr into  buffer at BufAddr.
          Read header & put NextAdr into CurrBlkAddr.}
         if not ReadABlock (CurrBlkAddr,BufAddr,GlobalHeader) then begin
            DoneWrBuffers(0, NotEof);
            CurrBlkAddr := DBLZERO;
         end
         else begin
            DoneWrBuffers(1, NotEof);
            CurrBlkAddr := GlobalHeader^.NextAdr;
         end;
   end {while <> DBLZERO};
   Got := AllocWrBuffers(0, Got);
   DoneWrBuffers(0, PseudoEof);
end {DumpFile};



procedure DumpDirTree(prefix: string);
{------------------------------------------------
{ Abstract:
{   Dumps an entire directory subtree to streamer (recursive)
{ Paratmers:
{   prefix - number of dashes to put before a subordinate idrectory
{            when its name is displayed.
{------------------------------------------------}
{Assume the FIB for a directory is in GlobalFIB}
var
   NextDiskAddr:DiskAddr;  {pts to current directory block}
   DirectoryBlk: ptrDiskBuffer;
   DirBlkSeg: SegmentNumber;
   i: integer;                     {for loop index}
begin {DumpDirTree}
   CreateSegment(DirBlkSeg, 1,1,1);
   DirectoryBlk := MakePtr (DirBlkSeg, 0, ptrDiskBuffer);
{  new(0,256,DirectoryBlk);    }
   NextDiskAddr := GlobalHeader^.NextAdr;
   while NextDiskAddr<>DBLZERO do 
      if not ReadABlock(NextDiskAddr, DirectoryBlk, GlobalHeader) then 
         NextDiskAddr := DBLZERO
      else begin {did read the next directory block}
         NextDiskAddr := GlobalHeader^.NextAdr;
         for i := 0 to FILESPERDIRBLK-1 do 
            with DirectoryBlk^.Entry[i] do 
               if InUse and not Deleted then begin
                  if not ReadABlock(ID, GlobalFIB, GlobalHeader) then
                     writeln ('Cannot process file.')
                  else if GlobalFIB^.FSData.FileType = DirFile then begin
                     writeln;
                     write (prefix, FileName);
                     DumpDirTree(Concat(prefix, '--'));
                  end
                  else begin
                     write('  ', FileName);
                     DumpFile(GlobalHeader^.NextAdr);
                  end;
               end;
      end;
   DecRefCount (DirBlkSeg);
{  dispose(DirectoryBlk);   }
end {DumpDirTree} ;




procedure FOWrite(Selections: ptrSelectEntry);
{------------------------------------------------
{ Abstract:
{   Writes one or more streamer tape partitions.
{ Parameters:
{   Selections - pts to the first in a chain of
{       SelectEntry blocks which indicate which disk
{       directories to put onto tape.
{------------------------------------------------}

label 30;
handler TSError(Status:StatBuf; Msg: TSMsg);
begin
   PrtStatus(Status, Msg);
   goto 30;
end;

var
   i, j: integer;
   Gotten: BufRange;
   Start, Now: TimeString;         
   Status: StatBuf;
   TSel: PtrSelectEntry;
   PartHdr: ptrPartBlk;
      
begin
   new(0, 256, GlobalFIB);
   new(0, 8,   GlobalHeader);

   GetTString(Start);  Writeln(Start);
   GrandTBlocks := 0;
   GrandTFiles  := 0;
   GrandTErrors := 0;

   TSel := Selections;
   while TSel <> nil do begin
      if not ReadABlock(TSel^.SelAddr, GlobalFIB, GlobalHeader) then
         writeln ('Cannot read root directory.')
      else begin
         for j := 1 to 3 do begin
            i := AllocWrBuffers(1, Gotten);
            if Gotten <> 1 then raise TSBufError(0);
            PartHdr := recast(BufPtrs[i], ptrPartBlk);
            PartHdr^.PartId := 'ParT';
            PartHdr^.PartName := TSel^.SelTpPart;
            PartHdr^.PartOriginal := TSel^.SelDir;
            PartHdr^.PartFSData := GlobalFib^.FSData;
            PartHdr^.PartNumInUse := GlobalFib^.NumBlksInUse;
            case j of                
            1,2:  DoneWrBuffers(1, NotEof);
            3:    DoneWrBuffers(1, PseudoEof);
            end;
         end;
         writeln;
         write('Write directory ', TSel^.SelDir, ' to tape partition ');
         PartHdr := recast(BufPtrs[i], ptrPartBlk);
         writeln(PartHdr^.PartName);
         FSAddToTitleLine(concat(STUTTitle, concat('   Putting ',
                        TSel^.SelDir)));
         NameStart := 1 + length(TSel^.SelDir) - PosC(TSel^.SelDir, '>');
         TotalBlocks := 0;  TotalFiles := 0;  TotalErrors := 0;

         DumpDirTree('--');

         Writeln;
         Writeln(TotalBlocks:8, ' Blocks',
                 TotalFiles:8,  ' Files',
                 TotalErrors:6, ' Disk Errors');
         GrandTBlocks := GrandTBlocks + TotalBlocks;
         GrandTFiles  := GrandTFiles  + TotalFiles;
         GrandTErrors := GrandTErrors + TotalErrors;
  
         i := AllocWrBuffers(1, Gotten);
         DoneWrBuffers(0, FileMark);       {Real FileMark at end partition}
      end;
      TSel := TSel^.SelNext;
   end;
   i := AllocWrBuffers(1, Gotten);
   DoneWrBuffers(0, EndTape);        {Finish the tape}
   FlushWrBuffers;                   {output all pending blocks}
30:
   GetTString(Now);  
   writeln;
   writeln('Started at ', Start, ' and finished at ', Now);
   Writeln(GrandTBlocks:8, ' Blocks',
           GrandTFiles:8,  ' Files',
           GrandTErrors:6, ' Disk Errors');
end.   
